/* edit-word-list.c
 *
 * Copyright 2023 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "crosswords-config.h"
#include <libipuz/libipuz.h>
#include <glib/gi18n-lib.h>
#include <adwaita.h>
#include "crosswords-enums.h"
#include "edit-grid.h"
#include "edit-word-list.h"
#include "play-grid.h"
#include "word-array-model.h"
#include "word-list.h"
#include "word-list-model.h"


enum
{
  WORD_ACTIVATED,
  N_SIGNALS
};

static guint obj_signals[N_SIGNALS] = { 0 };


struct _EditWordList
{
  GtkWidget parent_instance;

  GtkFilter *filter;
  IpuzClueDirection clue_direction;

  WordList *word_list;
  WordArrayModel *model;

  GtkWidget *suggested_words_prefs;
  GtkWidget *suggested_words_swindow;
  GtkWidget *suggested_words_list_view;

  GtkWidget *show_multi_switch_row;
  gboolean multi_check;

  gchar *across_clue_string;
  gchar *down_clue_string;
  guint across_pos;
  guint down_pos;
};


G_DEFINE_TYPE (EditWordList, edit_word_list, GTK_TYPE_WIDGET);


static void edit_word_list_init         (EditWordList      *self);
static void edit_word_list_class_init   (EditWordListClass *klass);
static void edit_word_list_dispose      (GObject           *object);
static void suggested_words_activate_cb (EditWordList      *self,
                                         guint              position);
static void show_multi_row_activated_cb (EditWordList      *self);

static void
setup_listitem_cb (GtkListItemFactory *factory,
                   GtkListItem        *list_item)
{
  GtkWidget *label;

  label = gtk_label_new (NULL);
  gtk_label_set_xalign (GTK_LABEL (label), 0.0);
  gtk_list_item_set_child (list_item, label);
}

static void
combine_word_and_enumeration_src (GString     *label_str,
                                  const gchar *word,
                                  const gchar *enumeration_src)
{
  const gchar *word_ptr;
  const gchar *ptr;

  g_assert (word);
  g_assert (enumeration_src);

  word_ptr = word;
  ptr = enumeration_src;
  while (ptr[0])
    {
      if (g_ascii_isdigit (ptr[0]))
        {
          guint64 num;
          gchar *endptr = NULL;
          num = g_ascii_strtoll (ptr, &endptr, 10);

          for (guint i = 0; i < num; i++)
            {
              g_assert (word_ptr[i]);
              g_string_append_unichar (label_str, word_ptr[i]);
            }
          word_ptr += num;
          ptr = endptr;
        }
      else
        {
          g_string_append_unichar (label_str, ptr[0]);
          ptr++;
        }
    }
}

static void
row_changed (WordListModelRow *row,
             GtkWidget        *label)
{
  GString *label_str = NULL;
  const gchar *enumeration_src;

  label_str = g_string_new (NULL);
  enumeration_src = word_list_model_row_get_enumeration_src (row);

  if (enumeration_src == NULL)
    {
      g_string_printf (label_str,
                       "%s — (%d)",
                       word_list_model_row_get_word (row),
                       word_list_model_row_get_priority (row));
    }
  else
    {
      combine_word_and_enumeration_src (label_str,
                                        word_list_model_row_get_word (row),
                                        enumeration_src);
      g_string_append_printf (label_str,
                              " - (%d)",
                              word_list_model_row_get_priority (row));
    }

  gtk_label_set_text (GTK_LABEL (label), label_str->str);
  g_string_free (label_str, TRUE);
}

static void
bind_listitem_cb (GtkListItemFactory *factory,
                  GtkListItem        *list_item)
{
  GtkWidget *label;
  WordListModelRow *row;

  label = gtk_list_item_get_child (list_item);
  row = (WordListModelRow *) gtk_list_item_get_item (list_item);

  g_signal_connect (row, "changed", G_CALLBACK (row_changed), label);
  row_changed (row, label);
}

static void
unbind_listitem_cb (GtkListItemFactory *factory,
                    GtkListItem        *list_item)
{
  GtkWidget *label;
  WordListModelRow *row;

  label = gtk_list_item_get_child (list_item);
  row = (WordListModelRow *) gtk_list_item_get_item (list_item);

  g_signal_handlers_disconnect_by_func (row, G_CALLBACK (row_changed), label);
}


gboolean
show_multi_filter (WordListModelRow *row,
                   EditWordList     *self)
{
  if (self->multi_check)
    return TRUE;

  return (word_list_model_row_get_enumeration_src (row) == NULL);
}

static void
edit_word_list_init (EditWordList *self)
{

  GListModel *filter_model;
  GListModel *selection_model;
  g_autoptr (GtkListItemFactory) factory = NULL;

  gtk_widget_init_template (GTK_WIDGET (self));
  self->multi_check = TRUE;

  self->word_list = word_list_new ();
  self->model = word_array_model_new (self->word_list);
  self->filter = (GtkFilter *)
    gtk_custom_filter_new ((GtkCustomFilterFunc) show_multi_filter,
                           self, NULL);
  factory = gtk_signal_list_item_factory_new ();
  g_signal_connect (factory, "setup", G_CALLBACK (setup_listitem_cb), NULL);
  g_signal_connect (factory, "bind", G_CALLBACK (bind_listitem_cb), NULL);
  g_signal_connect (factory, "unbind", G_CALLBACK (unbind_listitem_cb), NULL);

  filter_model = (GListModel *)
    gtk_filter_list_model_new (G_LIST_MODEL (self->model), self->filter);
  selection_model = (GListModel *) gtk_no_selection_new (filter_model);
  g_object_set (self->suggested_words_list_view,
                "model", selection_model,
                "factory", factory,
                NULL);
}

static void
edit_word_list_class_init (EditWordListClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->dispose = edit_word_list_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-word-list.ui");

  gtk_widget_class_bind_template_callback (widget_class, suggested_words_activate_cb);
  gtk_widget_class_bind_template_callback (widget_class, show_multi_row_activated_cb);

  gtk_widget_class_bind_template_child (widget_class, EditWordList, suggested_words_prefs);
  gtk_widget_class_bind_template_child (widget_class, EditWordList, suggested_words_swindow);
  gtk_widget_class_bind_template_child (widget_class, EditWordList, suggested_words_list_view);
  gtk_widget_class_bind_template_child (widget_class, EditWordList, show_multi_switch_row);

  obj_signals [WORD_ACTIVATED] =
    g_signal_new ("word-activated",
                  EDIT_TYPE_WORD_LIST,
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
                  0,
                  NULL, NULL,
                  NULL,
                  G_TYPE_NONE, 3,
                  IPUZ_TYPE_CLUE_DIRECTION,
                  G_TYPE_STRING,
                  G_TYPE_STRING);

  gtk_widget_class_set_css_name (widget_class, "wordlist");
  gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BIN_LAYOUT);
}

static void
edit_word_list_dispose (GObject *object)
{
  EditWordList *self;
  GtkWidget *child;

  self = EDIT_WORD_LIST (object);

  while ((child = gtk_widget_get_first_child (GTK_WIDGET (object))))
    gtk_widget_unparent (child);

  g_clear_object (&self->filter);
  g_clear_object (&self->word_list);
  g_clear_object (&self->model);
  g_clear_pointer (&self->across_clue_string, g_free);
  g_clear_pointer (&self->down_clue_string, g_free);

  G_OBJECT_CLASS (edit_word_list_parent_class)->dispose (object);
}

static void
word_activate (EditWordList      *self,
               guint              position,
               IpuzClueDirection  direction)
{
  GListModel *model;
  WordListModelRow *row;
  const gchar *word;
  const gchar *enumeration_src;

  model = (GListModel *) gtk_list_view_get_model (GTK_LIST_VIEW (self->suggested_words_list_view));

  row = g_list_model_get_item (model, position);
  word = word_list_model_row_get_word (row);
  enumeration_src = word_list_model_row_get_enumeration_src (row);
  g_assert (word);

  g_signal_emit (self, obj_signals [WORD_ACTIVATED], 0, direction, word, enumeration_src);
}

static void
suggested_words_activate_cb (EditWordList *self,
                             guint         position)
{
  IpuzClueDirection dir;

  dir = self->clue_direction;
  g_assert (dir == IPUZ_CLUE_DIRECTION_ACROSS ||
            dir == IPUZ_CLUE_DIRECTION_DOWN);
  word_activate (self, position, dir);
}

static void
show_multi_row_activated_cb (EditWordList *self)
{
  gboolean multi_check;

  multi_check = adw_switch_row_get_active (ADW_SWITCH_ROW (self->show_multi_switch_row));
  if (multi_check != self->multi_check)
    {
      GtkFilterChange change;
      self->multi_check = multi_check;

      if (multi_check)
        change = GTK_FILTER_CHANGE_LESS_STRICT;
      else
        change = GTK_FILTER_CHANGE_MORE_STRICT;

      gtk_filter_changed (self->filter, change);
    }
}


static void
calculate_clue_word_pos (GridState          *state,
                         IpuzClueDirection   dir,
                         gchar             **word,
                         guint              *pos)
{
  IpuzClue *clue;
  IpuzCellCoordArray *arr;
  IpuzCellCoord coord;
  guint i = 0;
  GString *str;
  IpuzGuesses *guesses;

  g_assert (state);
  g_assert (word);
  g_assert (pos);

  g_clear_pointer (word, g_free);
  *pos = 0;

  clue = ipuz_clues_find_clue_by_coord (IPUZ_CLUES (state->xword),
                                        dir,
                                        &state->cursor);
  if (clue == NULL)
    return;

  str = g_string_new (NULL);
  arr = ipuz_clue_get_coords (clue);
  guesses = ipuz_grid_get_guesses (IPUZ_GRID (state->xword));

  while (ipuz_cell_coord_array_index (arr, i++, &coord))
    {
      IpuzCell *cell;

      /* FIXME(text): This doesn't work for
        * graphemes with multiple code points.
        * See devel-docs/basic-types.html#grid-words
        * for more information. */
      /* Use strlen to account for rebuses correctly. */
      if (ipuz_cell_coord_equal (&state->cursor, &coord))
        *pos = g_utf8_strlen (str->str, -1);

      cell = ipuz_grid_get_cell (IPUZ_GRID (state->xword), &coord);
      if (ipuz_cell_get_initial_val (cell))
        g_string_append (str, ipuz_cell_get_initial_val (cell));
      else if (ipuz_cell_get_solution (cell))
        g_string_append (str, ipuz_cell_get_solution (cell));
      else if (guesses) /* pencil markings */
        {
          const gchar *guess;

          guess = ipuz_guesses_get_guess (guesses, &coord);
          if (guess)
            g_string_append (str, guess);
          else
            g_string_append (str, "?");
        }
      else
        g_string_append (str, "?");
    }

  *word = g_string_free_and_steal (str);
}

static void
update_model (EditWordList *self)
{
  IpuzClueDirection dir;

  dir = self->clue_direction;
  g_assert (dir == IPUZ_CLUE_DIRECTION_ACROSS ||
            dir == IPUZ_CLUE_DIRECTION_DOWN);

  g_autoptr (IpuzCharset) intersecting_chars = NULL;
  g_autoptr (WordArray) across_array = NULL;
  g_autoptr (WordArray) down_array = NULL;

  word_list_find_intersection (self->word_list,
                               self->across_clue_string,
                               self->across_pos,
                               self->down_clue_string,
                               self->down_pos,
                               &intersecting_chars,
                               &across_array,
                               &down_array);

  if (dir == IPUZ_CLUE_DIRECTION_ACROSS)
    word_array_model_set_array (self->model, across_array);
  else
    word_array_model_set_array (self->model, down_array);

  gtk_filter_changed (self->filter, GTK_FILTER_CHANGE_DIFFERENT);
}

void
edit_word_list_update (EditWordList     *self,
                       GridState        *edit_state,
                       WordListResource *resource)
{
  g_return_if_fail (EDIT_IS_WORD_LIST (self));

  AdwPreferencesGroup *prefs_group;

  prefs_group = ADW_PREFERENCES_GROUP (self->suggested_words_prefs);
  self->across_pos = 0;
  self->down_pos = 0;
  self->clue_direction = edit_state->clue.direction;
  g_clear_pointer (&self->across_clue_string, g_free);
  g_clear_pointer (&self->down_clue_string, g_free);

  word_list_set_resource (self->word_list, resource);
  if (self->clue_direction == IPUZ_CLUE_DIRECTION_NONE)
    {
      word_array_model_set_array (self->model, NULL);
      adw_preferences_group_set_title (prefs_group, " ");
      return;
    }

  const gchar *title = ipuz_clue_direction_to_string (self->clue_direction);
  adw_preferences_group_set_title (prefs_group, title);

  if (!IPUZ_IS_ACROSTIC (edit_state->xword))
    {
      calculate_clue_word_pos (edit_state,
                               IPUZ_CLUE_DIRECTION_ACROSS,
                               &self->across_clue_string,
                               &self->across_pos);
      calculate_clue_word_pos (edit_state,
                               IPUZ_CLUE_DIRECTION_DOWN,
                               &self->down_clue_string,
                               &self->down_pos);
    }
  else
    {
      calculate_clue_word_pos (edit_state,
                               IPUZ_CLUE_DIRECTION_CLUES,
                               &self->across_clue_string,
                               &self->across_pos);

    }
  update_model (self);
}